@@ -4,18 +4,25 @@
     is capable of handling the signatures.
 
 """
+import datetime
 import hashlib
 import logging
 import re
-import warnings
+from typing import Callable
+from typing import List
+from typing import Optional
+from typing import Set
+from typing import Tuple
+from typing import TYPE_CHECKING
+from typing import Union
 
-from typing import List, Set
-# See https://github.com/pyca/cryptography/issues/4275
-from cryptography import x509  # type: ignore
+from cryptography import x509
 from cryptography.exceptions import InvalidSignature
 from cryptography.exceptions import UnsupportedAlgorithm
 from cryptography.hazmat.backends import default_backend
+from cryptography.hazmat.primitives import hashes
 from cryptography.hazmat.primitives.asymmetric import ec
+from cryptography.hazmat.primitives.asymmetric.dsa import DSAPublicKey
 from cryptography.hazmat.primitives.asymmetric.ec import ECDSA
 from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicKey
 from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
@@ -23,10 +30,10 @@ from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
 from cryptography.hazmat.primitives.serialization import Encoding
 from cryptography.hazmat.primitives.serialization import NoEncryption
 from cryptography.hazmat.primitives.serialization import PrivateFormat
+import josepy
 from OpenSSL import crypto
-from OpenSSL import SSL  # type: ignore
+from OpenSSL import SSL
 import pyrfc3339
-import zope.component
 
 from acme import crypto_util as acme_crypto_util
 from certbot import errors
@@ -34,6 +41,11 @@ from certbot import interfaces
 from certbot import util
 from certbot.compat import os
 
+# Cryptography ed448 and ed25519 modules do not exist on oldest tests
+if TYPE_CHECKING:
+    from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PublicKey
+    from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
+
 logger = logging.getLogger(__name__)
 
 
@@ -86,41 +98,7 @@ def generate_key(key_size: int, key_dir: str, key_type: str = "rsa",
     return util.Key(key_path, key_pem)
 
 
-# TODO: Remove this call once zope dependencies are removed from Certbot.
-def init_save_key(key_size, key_dir, key_type="rsa", elliptic_curve="secp256r1",
-                  keyname="key-certbot.pem"):
-    """Initializes and saves a privkey.
-
-    Inits key and saves it in PEM format on the filesystem.
-
-    .. note:: keyname is the attempted filename, it may be different if a file
-        already exists at the path.
-
-    .. deprecated:: 1.16.0
-       Use :func:`generate_key` instead.
-
-    :param int key_size: key size in bits if key size is rsa.
-    :param str key_dir: Key save directory.
-    :param str key_type: Key Type [rsa, ecdsa]
-    :param str elliptic_curve: Name of the elliptic curve if key type is ecdsa.
-    :param str keyname: Filename of key
-
-    :returns: Key
-    :rtype: :class:`certbot.util.Key`
-
-    :raises ValueError: If unable to generate the key given key_size.
-
-    """
-    warnings.warn("certbot.crypto_util.init_save_key is deprecated, please use "
-                  "certbot.crypto_util.generate_key instead.", DeprecationWarning)
-
-    config = zope.component.getUtility(interfaces.IConfig)
-
-    return generate_key(key_size, key_dir, key_type=key_type, elliptic_curve=elliptic_curve,
-                        keyname=keyname, strict_permissions=config.strict_permissions)
-
-
-def generate_csr(privkey: util.Key, names: Set[str], path: str,
+def generate_csr(privkey: util.Key, names: Union[List[str], Set[str]], path: str,
                  must_staple: bool = False, strict_permissions: bool = True) -> util.CSR:
     """Initialize a CSR with the given private key.
 
@@ -128,7 +106,7 @@ def generate_csr(privkey: util.Key, names: Set[str], path: str,
     :type privkey: :class:`certbot.util.Key`
     :param set names: `str` names to include in the CSR
     :param str path: Certificate save directory.
-    :param bool must_staple: If true, include the TLS Feature extension "OCSP Must Staple"
+    :param bool must_staple: If true, include the TLS Feature extension "OCSP Must-Staple"
     :param bool strict_permissions: If true and path exists, an exception is raised if
         the directory doesn't have 0755 permissions or isn't owned by the current user.
 
@@ -150,44 +128,17 @@ def generate_csr(privkey: util.Key, names: Set[str], path: str,
     return util.CSR(csr_filename, csr_pem, "pem")
 
 
-# TODO: Remove this call once zope dependencies are removed from Certbot.
-def init_save_csr(privkey, names, path):
-    """Initialize a CSR with the given private key.
-
-    .. deprecated:: 1.16.0
-       Use :func:`generate_csr` instead.
-
-    :param privkey: Key to include in the CSR
-    :type privkey: :class:`certbot.util.Key`
-
-    :param set names: `str` names to include in the CSR
-
-    :param str path: Certificate save directory.
-
-    :returns: CSR
-    :rtype: :class:`certbot.util.CSR`
-
-    """
-    warnings.warn("certbot.crypto_util.init_save_csr is deprecated, please use "
-                  "certbot.crypto_util.generate_csr instead.", DeprecationWarning)
-
-    config = zope.component.getUtility(interfaces.IConfig)
-
-    return generate_csr(privkey, names, path, must_staple=config.must_staple,
-                        strict_permissions=config.strict_permissions)
-
-
 # WARNING: the csr and private key file are possible attack vectors for TOCTOU
 # We should either...
 # A. Do more checks to verify that the CSR is trusted/valid
 # B. Audit the parsing code for vulnerabilities
 
-def valid_csr(csr):
+def valid_csr(csr: bytes) -> bool:
     """Validate CSR.
 
     Check if `csr` is a valid CSR for the given domains.
 
-    :param str csr: CSR in PEM.
+    :param bytes csr: CSR in PEM.
 
     :returns: Validity of CSR.
     :rtype: bool
@@ -202,11 +153,11 @@ def valid_csr(csr):
         return False
 
 
-def csr_matches_pubkey(csr, privkey):
+def csr_matches_pubkey(csr: bytes, privkey: bytes) -> bool:
     """Does private key correspond to the subject public key in the CSR?
 
-    :param str csr: CSR in PEM.
-    :param str privkey: Private key file contents (PEM)
+    :param bytes csr: CSR in PEM.
+    :param bytes privkey: Private key file contents (PEM)
 
     :returns: Correspondence of private key to CSR subject public key.
     :rtype: bool
@@ -222,11 +173,11 @@ def csr_matches_pubkey(csr, privkey):
         return False
 
 
-def import_csr_file(csrfile, data):
+def import_csr_file(csrfile: str, data: bytes) -> Tuple[int, util.CSR, List[str]]:
     """Import a CSR file, which can be either PEM or DER.
 
     :param str csrfile: CSR filename
-    :param str data: contents of the CSR file
+    :param bytes data: contents of the CSR file
 
     :returns: (`crypto.FILETYPE_PEM`,
                util.CSR object representing the CSR,
@@ -251,12 +202,13 @@ def import_csr_file(csrfile, data):
     return PEM, util.CSR(file=csrfile, data=data_pem, form="pem"), domains
 
 
-def make_key(bits=1024, key_type="rsa", elliptic_curve=None):
+def make_key(bits: int = 1024, key_type: str = "rsa",
+             elliptic_curve: Optional[str] = None) -> bytes:
     """Generate PEM encoded RSA|EC key.
 
     :param int bits: Number of bits if key_type=rsa. At least 1024 for RSA.
-
-    :param str ec_curve: The elliptic curve to use.
+    :param str key_type: The type of key to generate, but be rsa or ecdsa
+    :param str elliptic_curve: The elliptic curve to use.
 
     :returns: new RSA or ECDSA key in PEM form with specified number of bits
               or of type ec_curve when key_type ecdsa is used.
@@ -269,11 +221,16 @@ def make_key(bits=1024, key_type="rsa", elliptic_curve=None):
         key = crypto.PKey()
         key.generate_key(crypto.TYPE_RSA, bits)
     elif key_type == 'ecdsa':
+        if not elliptic_curve:
+            raise errors.Error("When key_type == ecdsa, elliptic_curve must be set.")
         try:
             name = elliptic_curve.upper()
             if name in ('SECP256R1', 'SECP384R1', 'SECP521R1'):
+                curve = getattr(ec, elliptic_curve.upper())
+                if not curve:
+                    raise errors.Error(f"Invalid curve type: {elliptic_curve}")
                 _key = ec.generate_private_key(
-                    curve=getattr(ec, elliptic_curve.upper(), None)(),
+                    curve=curve(),
                     backend=default_backend()
                 )
             else:
@@ -282,7 +239,11 @@ def make_key(bits=1024, key_type="rsa", elliptic_curve=None):
             raise errors.Error("Unsupported elliptic curve: {}".format(elliptic_curve))
         except UnsupportedAlgorithm as e:
             raise e from errors.Error(str(e))
-        _key_pem = _key.private_bytes(
+        # This type ignore directive is required due to an outdated version of types-cryptography.
+        # It can be removed once package types-pyOpenSSL depends on cryptography instead of
+        # types-cryptography and so types-cryptography is not installed anymore.
+        # See https://github.com/python/typeshed/issues/5618
+        _key_pem = _key.private_bytes(  # type: ignore
             encoding=Encoding.PEM,
             format=PrivateFormat.TraditionalOpenSSL,
             encryption_algorithm=NoEncryption()
@@ -293,7 +254,7 @@ def make_key(bits=1024, key_type="rsa", elliptic_curve=None):
     return crypto.dump_privatekey(crypto.FILETYPE_PEM, key)
 
 
-def valid_privkey(privkey):
+def valid_privkey(privkey: str) -> bool:
     """Is valid RSA private key?
 
     :param str privkey: Private key file contents in PEM
@@ -309,7 +270,7 @@ def valid_privkey(privkey):
         return False
 
 
-def verify_renewable_cert(renewable_cert):
+def verify_renewable_cert(renewable_cert: interfaces.RenewableCert) -> None:
     """For checking that your certs were not corrupted on disk.
 
     Several things are checked:
@@ -327,7 +288,7 @@ def verify_renewable_cert(renewable_cert):
     verify_cert_matches_priv_key(renewable_cert.cert_path, renewable_cert.key_path)
 
 
-def verify_renewable_cert_sig(renewable_cert):
+def verify_renewable_cert_sig(renewable_cert: interfaces.RenewableCert) -> None:
     """Verifies the signature of a RenewableCert object.
 
     :param renewable_cert: cert to verify
@@ -341,9 +302,8 @@ def verify_renewable_cert_sig(renewable_cert):
         with open(renewable_cert.cert_path, 'rb') as cert_file:
             cert = x509.load_pem_x509_certificate(cert_file.read(), default_backend())
         pk = chain.public_key()
-        with warnings.catch_warnings():
-            verify_signed_payload(pk, cert.signature, cert.tbs_certificate_bytes,
-                                  cert.signature_hash_algorithm)
+        verify_signed_payload(pk, cert.signature, cert.tbs_certificate_bytes,
+                                cert.signature_hash_algorithm)
     except (IOError, ValueError, InvalidSignature) as e:
         error_str = "verifying the signature of the certificate located at {0} has failed. \
                 Details: {1}".format(renewable_cert.cert_path, e)
@@ -351,37 +311,33 @@ def verify_renewable_cert_sig(renewable_cert):
         raise errors.Error(error_str)
 
 
-def verify_signed_payload(public_key, signature, payload, signature_hash_algorithm):
+def verify_signed_payload(public_key: Union[DSAPublicKey, 'Ed25519PublicKey', 'Ed448PublicKey',
+                                            EllipticCurvePublicKey, RSAPublicKey],
+                          signature: bytes, payload: bytes,
+                          signature_hash_algorithm: hashes.HashAlgorithm) -> None:
     """Check the signature of a payload.
 
     :param RSAPublicKey/EllipticCurvePublicKey public_key: the public_key to check signature
     :param bytes signature: the signature bytes
     :param bytes payload: the payload bytes
-    :param cryptography.hazmat.primitives.hashes.HashAlgorithm \
-           signature_hash_algorithm: algorithm used to hash the payload
+    :param hashes.HashAlgorithm signature_hash_algorithm: algorithm used to hash the payload
 
     :raises InvalidSignature: If signature verification fails.
     :raises errors.Error: If public key type is not supported
     """
-    with warnings.catch_warnings():
-        warnings.simplefilter("ignore")
-        if isinstance(public_key, RSAPublicKey):
-            verifier = public_key.verifier(
-                signature, PKCS1v15(), signature_hash_algorithm
-            )
-            verifier.update(payload)
-            verifier.verify()
-        elif isinstance(public_key, EllipticCurvePublicKey):
-            verifier = public_key.verifier(
-                signature, ECDSA(signature_hash_algorithm)
-            )
-            verifier.update(payload)
-            verifier.verify()
-        else:
-            raise errors.Error("Unsupported public key type")
+    if isinstance(public_key, RSAPublicKey):
+        public_key.verify(
+            signature, payload, PKCS1v15(), signature_hash_algorithm
+        )
+    elif isinstance(public_key, EllipticCurvePublicKey):
+        public_key.verify(
+            signature, payload, ECDSA(signature_hash_algorithm)
+        )
+    else:
+        raise errors.Error("Unsupported public key type.")
 
 
-def verify_cert_matches_priv_key(cert_path, key_path):
+def verify_cert_matches_priv_key(cert_path: str, key_path: str) -> None:
     """ Verifies that the private key and cert match.
 
     :param str cert_path: path to a cert in PEM format
@@ -403,7 +359,7 @@ def verify_cert_matches_priv_key(cert_path, key_path):
         raise errors.Error(error_str)
 
 
-def verify_fullchain(renewable_cert):
+def verify_fullchain(renewable_cert: interfaces.RenewableCert) -> None:
     """ Verifies that fullchain is indeed cert concatenated with chain.
 
     :param renewable_cert: cert to verify
@@ -430,7 +386,7 @@ def verify_fullchain(renewable_cert):
         raise e
 
 
-def pyopenssl_load_certificate(data):
+def pyopenssl_load_certificate(data: bytes) -> Tuple[crypto.X509, int]:
     """Load PEM/DER certificate.
 
     :raises errors.Error:
@@ -448,8 +404,9 @@ def pyopenssl_load_certificate(data):
         str(error) for error in openssl_errors)))
 
 
-def _load_cert_or_req(cert_or_req_str, load_func,
-                      typ=crypto.FILETYPE_PEM):
+def _load_cert_or_req(cert_or_req_str: bytes,
+                      load_func: Callable[[int, bytes], Union[crypto.X509, crypto.X509Req]],
+                      typ: int = crypto.FILETYPE_PEM) -> Union[crypto.X509, crypto.X509Req]:
     try:
         return load_func(typ, cert_or_req_str)
     except crypto.Error as err:
@@ -458,14 +415,16 @@ def _load_cert_or_req(cert_or_req_str, load_func,
         raise
 
 
-def _get_sans_from_cert_or_req(cert_or_req_str, load_func,
-                               typ=crypto.FILETYPE_PEM):
+def _get_sans_from_cert_or_req(cert_or_req_str: bytes,
+                               load_func: Callable[[int, bytes], Union[crypto.X509,
+                                                                       crypto.X509Req]],
+                               typ: int = crypto.FILETYPE_PEM) -> List[str]:
     # pylint: disable=protected-access
     return acme_crypto_util._pyopenssl_cert_or_req_san(_load_cert_or_req(
         cert_or_req_str, load_func, typ))
 
 
-def get_sans_from_cert(cert, typ=crypto.FILETYPE_PEM):
+def get_sans_from_cert(cert: bytes, typ: int = crypto.FILETYPE_PEM) -> List[str]:
     """Get a list of Subject Alternative Names from a certificate.
 
     :param str cert: Certificate (encoded).
@@ -479,17 +438,21 @@ def get_sans_from_cert(cert, typ=crypto.FILETYPE_PEM):
         cert, crypto.load_certificate, typ)
 
 
-def _get_names_from_cert_or_req(cert_or_req, load_func, typ):
+def _get_names_from_cert_or_req(cert_or_req: bytes,
+                                load_func: Callable[[int, bytes], Union[crypto.X509,
+                                                                        crypto.X509Req]],
+                                typ: int) -> List[str]:
     loaded_cert_or_req = _load_cert_or_req(cert_or_req, load_func, typ)
     return _get_names_from_loaded_cert_or_req(loaded_cert_or_req)
 
 
-def _get_names_from_loaded_cert_or_req(loaded_cert_or_req):
+def _get_names_from_loaded_cert_or_req(loaded_cert_or_req: Union[crypto.X509, crypto.X509Req]
+                                       ) -> List[str]:
     # pylint: disable=protected-access
     return acme_crypto_util._pyopenssl_cert_or_req_all_names(loaded_cert_or_req)
 
 
-def get_names_from_cert(csr, typ=crypto.FILETYPE_PEM):
+def get_names_from_cert(cert: bytes, typ: int = crypto.FILETYPE_PEM) -> List[str]:
     """Get a list of domains from a cert, including the CN if it is set.
 
     :param str cert: Certificate (encoded).
@@ -500,13 +463,13 @@ def get_names_from_cert(csr, typ=crypto.FILETYPE_PEM):
 
     """
     return _get_names_from_cert_or_req(
-        csr, crypto.load_certificate, typ)
+        cert, crypto.load_certificate, typ)
 
 
-def get_names_from_req(csr: str, typ: int = crypto.FILETYPE_PEM) -> List[str]:
+def get_names_from_req(csr: bytes, typ: int = crypto.FILETYPE_PEM) -> List[str]:
     """Get a list of domains from a CSR, including the CN if it is set.
 
-    :param str cert: CSR (encoded).
+    :param str csr: CSR (encoded).
     :param typ: `crypto.FILETYPE_PEM` or `crypto.FILETYPE_ASN1`
     :returns: A list of domain names.
     :rtype: list
@@ -515,7 +478,8 @@ def get_names_from_req(csr: str, typ: int = crypto.FILETYPE_PEM) -> List[str]:
     return _get_names_from_cert_or_req(csr, crypto.load_certificate_request, typ)
 
 
-def dump_pyopenssl_chain(chain, filetype=crypto.FILETYPE_PEM):
+def dump_pyopenssl_chain(chain: Union[List[crypto.X509], List[josepy.ComparableX509]],
+                         filetype: int = crypto.FILETYPE_PEM) -> bytes:
     """Dump certificate chain into a bundle.
 
     :param list chain: List of `crypto.X509` (or wrapped in
@@ -527,7 +491,7 @@ def dump_pyopenssl_chain(chain, filetype=crypto.FILETYPE_PEM):
     return acme_crypto_util.dump_pyopenssl_chain(chain, filetype)
 
 
-def notBefore(cert_path):
+def notBefore(cert_path: str) -> datetime.datetime:
     """When does the cert at cert_path start being valid?
 
     :param str cert_path: path to a cert in PEM format
@@ -539,7 +503,7 @@ def notBefore(cert_path):
     return _notAfterBefore(cert_path, crypto.X509.get_notBefore)
 
 
-def notAfter(cert_path):
+def notAfter(cert_path: str) -> datetime.datetime:
     """When does the cert at cert_path stop being valid?
 
     :param str cert_path: path to a cert in PEM format
@@ -551,7 +515,8 @@ def notAfter(cert_path):
     return _notAfterBefore(cert_path, crypto.X509.get_notAfter)
 
 
-def _notAfterBefore(cert_path, method):
+def _notAfterBefore(cert_path: str,
+                    method: Callable[[crypto.X509], Optional[bytes]]) -> datetime.datetime:
     """Internal helper function for finding notbefore/notafter.
 
     :param str cert_path: path to a cert in PEM format
@@ -567,6 +532,8 @@ def _notAfterBefore(cert_path, method):
         x509 = crypto.load_certificate(crypto.FILETYPE_PEM, f.read())
     # pyopenssl always returns bytes
     timestamp = method(x509)
+    if not timestamp:
+        raise errors.Error("Error while invoking timestamp method, None has been returned.")
     reformatted_timestamp = [timestamp[0:4], b"-", timestamp[4:6], b"-",
                              timestamp[6:8], b"T", timestamp[8:10], b":",
                              timestamp[10:12], b":", timestamp[12:]]
@@ -576,7 +543,7 @@ def _notAfterBefore(cert_path, method):
     return pyrfc3339.parse(timestamp_str)
 
 
-def sha256sum(filename):
+def sha256sum(filename: str) -> str:
     """Compute a sha256sum of a file.
 
     NB: In given file, platform specific newlines characters will be converted
@@ -603,7 +570,7 @@ CERT_PEM_REGEX = re.compile(
 )
 
 
-def cert_and_chain_from_fullchain(fullchain_pem):
+def cert_and_chain_from_fullchain(fullchain_pem: str) -> Tuple[str, str]:
     """Split fullchain_pem into cert_pem and chain_pem
 
     :param str fullchain_pem: concatenated cert + chain
@@ -631,7 +598,7 @@ def cert_and_chain_from_fullchain(fullchain_pem):
     return (certs_normalized[0], "".join(certs_normalized[1:]))
 
 
-def get_serial_from_cert(cert_path):
+def get_serial_from_cert(cert_path: str) -> int:
     """Retrieve the serial number of a certificate from certificate path
 
     :param str cert_path: path to a cert in PEM format
@@ -645,7 +612,8 @@ def get_serial_from_cert(cert_path):
     return x509.get_serial_number()
 
 
-def find_chain_with_issuer(fullchains, issuer_cn, warn_on_no_match=False):
+def find_chain_with_issuer(fullchains: List[str], issuer_cn: str,
+                           warn_on_no_match: bool = False) -> str:
     """Chooses the first certificate chain from fullchains whose topmost
     intermediate has an Issuer Common Name matching issuer_cn (in other words
     the first chain which chains to a root whose name matches issuer_cn).
